using namespace std;

#include <iostream>
#include <assert.h>
#include <unistd.h>
#include <stdlib.h>

extern "C" {
#include <X11/Xlib.h>
#include <X11/Xutil.h>
}

#include "boardstruc.h"
#include "boardprint.h"
#include "players.h"

const float ONE_OVER_ROOT_3 = 0.5773502691;
const float VH_RATIO = 16.0 / (24.0 * ONE_OVER_ROOT_3);

const int MIN_WINDOW_WIDTH = 400;
const int MIN_WINDOW_HEIGHT = int(MIN_WINDOW_WIDTH * VH_RATIO);

const int WINDOW_OFFSET_X = 10;  // offsets of window from corner of screen
const int WINDOW_OFFSET_Y = 10;

Display *display;
Window win;
Colormap colormap;

GC darkGreyGC, greyGC, lightGreyGC;
GC lightRedGC, dimmedRedGC, redGC;
GC lightGreenGC, dimmedGreenGC, greenGC;
GC blackGC;
const char darkGreyName[] = "azure4";
const char greyName[] = "azure3";
const char lightGreyName[] = "azure2";
const char lightRedName[] = "salmon1";
const char dimmedRedName[] = "salmon2";
const char redName[] = "salmon3";
const char lightGreenName[] = "OliveDrab2";
const char dimmedGreenName[] = "OliveDrab3";
const char greenName[] = "OliveDrab4";
const char blackName[] = "grey38";

int verticalSkip;
int inset;
int height;
int width;
int size;
int marbleRadius;
int holeRadius;

// Draw a circle centered at (x, y) with radius r (measurements in pixels)
void DrawCircle(const GC &gc, int x, int y, int r)
{
  XFillArc(display, win, gc, x-r-1, y-r-1, 2*r+1, 2*r+1, 0, 360*64);
  XFlush(display);
}

struct Point {
  int v;
  int h;
};

Point Translate(int x, int y)
{
  Point result;
  
  float xx = x * ONE_OVER_ROOT_3 * verticalSkip;
  float yy = y * verticalSkip;
  
  xx = xx + width/2;
  yy = size - yy + inset;
        
  result.v = int(yy);
  result.h = int(xx);
        
  return result;
}

void DrawOneOrientation(MarblePosType posList[17][16])
{
  for (int i = 1; i < 16; i++) {
    int pos1 = posList[i][0];
    int pos2 = posList[i][posPerLine[i]-1];
    Point p1 = Translate(xPos[pos1], yPos[pos1]);
    Point p2 = Translate(xPos[pos2], yPos[pos2]);
    XDrawLine(display, win, blackGC, p1.h, p1.v, p2.h, p2.v);
  }
}

void DrawOnePos(MarblePosType i)
{
  Point p = Translate(xPos[i], yPos[i]);
  switch (posColor[i]) {
  case LG:
    DrawCircle(lightGreyGC, p.h, p.v, holeRadius);
    break;
  case MG:
    DrawCircle(greyGC, p.h, p.v, holeRadius);
    break;
  case DG:
    DrawCircle(darkGreyGC, p.h, p.v, holeRadius);
  }
}

void DrawPositions(void)
{
  for (int i = 0; i < 121; i++)
    DrawOnePos(i);
}

void DrawOneMarble(MarblePosType pos, GC gc)
{
  Point p = Translate(xPos[pos], yPos[pos]);
  DrawCircle(gc, p.h, p.v, marbleRadius);
}

void DrawMarbles(void)
{
  int i;

  for (i = 0; i < 10; i++)
    DrawOneMarble(currBoard[0][i], redGC);

  for (i = 0; i < 10; i++)
    DrawOneMarble(currBoard[1][i], greenGC);
}

void DrawBoard()
{
  size = height;  
  inset = int(size * 0.05);
  size = size - 2 * inset;
  verticalSkip = int(size / 16.0);
  marbleRadius = int(verticalSkip * 0.4);
  holeRadius = int(verticalSkip * 0.4);
  
  DrawOneOrientation(horiz);
  DrawOneOrientation(slash);
  DrawOneOrientation(bSlash);

  DrawPositions();
  
  DrawMarbles();

  XFlush(display);
}

const int NUM_BLINKS = 3;
const unsigned long SLEEP_TIME = 75000;

void AnimateFrom(GC &marbleColor, GC &dimmedColor, int pos)
{
  for (int j = 0; j < NUM_BLINKS; j++) {
    DrawOnePos(pos);
    usleep(SLEEP_TIME);
    DrawOneMarble(pos, marbleColor);
    usleep(SLEEP_TIME);
  }
  DrawOneMarble(pos, dimmedColor);
}

void AnimatePath(GC &pathColor, MarblePosType path[121], int pathLen)
{
  for (int i = 0; i < pathLen-1; i++) {
    DrawOneMarble(path[i], pathColor);
    usleep(SLEEP_TIME);
  }
}

void AnimateTo(GC &marbleColor, int pos)
{
  for (int j = 0; j < NUM_BLINKS; j++) {
    DrawOnePos(pos);
    usleep(SLEEP_TIME);
    DrawOneMarble(pos, marbleColor);
    usleep(SLEEP_TIME);
  }
}

void ClearFrom(int pos)
{
  DrawOnePos(pos);
}

void ClearPath(MarblePosType path[121], int pathLen)
{
  for (int i = 0; i < pathLen-1; i++) {
    DrawOnePos(path[i]);
    usleep(SLEEP_TIME);
  }
}

void DisplayMove(int player, MarblePosType from, MarblePosType to,
		 MarblePosType path[121], int pathLen)
{
  if (player == 1) {
    AnimateFrom(redGC, dimmedRedGC, from);
    AnimatePath(lightRedGC, path, pathLen);
    AnimateTo(redGC, to);
    ClearFrom(from);
    ClearPath(path, pathLen);
  }
  else {
    AnimateFrom(greenGC, dimmedGreenGC, from);
    AnimatePath(lightGreenGC, path, pathLen);
    AnimateTo(greenGC, to);
    ClearFrom(from);
    ClearPath(path, pathLen);
  }
}

void InitXWindow(int argc, char *argv[])
{
  char *progname = argv[0];

  XSizeHints *size_hints = XAllocSizeHints();
  assert(size_hints);

  XWMHints *wm_hints = XAllocWMHints();
  assert(wm_hints);

  XClassHint *class_hints = XAllocClassHint();
  assert(class_hints);

  assert(display = XOpenDisplay(NULL));

  int screen_num = DefaultScreen(display);
  colormap = DefaultColormap(display, screen_num);

  const int HEIGHT_SLACK = 48;  // allow for window frame, etc.
  const int MAX_WINDOW_HEIGHT = DisplayHeight(display, screen_num)
    - HEIGHT_SLACK;
  const int MAX_WINDOW_WIDTH = int(MAX_WINDOW_HEIGHT / VH_RATIO);

  win = XCreateSimpleWindow(display, RootWindow(display, screen_num),
			    WINDOW_OFFSET_X, WINDOW_OFFSET_Y,
			    MAX_WINDOW_WIDTH, MAX_WINDOW_HEIGHT, 0,
			    BlackPixel(display, screen_num),
			    WhitePixel(display, screen_num));

  // specify a non-resizable window (makes things simpler)
  size_hints->flags = PPosition | PSize | PMinSize | PMaxSize | PAspect;
  size_hints->min_width = MIN_WINDOW_WIDTH;
  size_hints->min_height = MIN_WINDOW_HEIGHT;
  size_hints->max_width = MAX_WINDOW_WIDTH;
  size_hints->max_height = MAX_WINDOW_HEIGHT;
  size_hints->min_aspect.x = MIN_WINDOW_WIDTH;
  size_hints->min_aspect.y = MIN_WINDOW_HEIGHT;
  size_hints->max_aspect.x = MIN_WINDOW_WIDTH;
  size_hints->max_aspect.y = MIN_WINDOW_HEIGHT;

  char *window_name = "Chinese Checkers";
  XTextProperty windowName;
  assert(XStringListToTextProperty(&window_name, 1, &windowName));

  wm_hints->initial_state = NormalState;
  wm_hints->input = True;
  wm_hints->flags = StateHint | InputHint;

  class_hints->res_name = progname;
  class_hints->res_class = "Basicwin";
  
  XSetWMProperties(display, win, &windowName, &windowName, argv, argc,
		   size_hints, wm_hints, class_hints);
  XSelectInput(display, win, KeyPressMask | ButtonPressMask
	       | ExposureMask | StructureNotifyMask);

  // set up two graphics contexts: one for drawing, one for erasing
  unsigned long valuemask = 0;
  XGCValues values;
  XColor availColor, exactRGB;

  darkGreyGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, darkGreyName, &availColor, &exactRGB);
  XSetForeground(display, darkGreyGC, availColor.pixel);
  XSetFillStyle(display, darkGreyGC, FillSolid);

  greyGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, greyName, &availColor, &exactRGB);
  XSetForeground(display, greyGC, availColor.pixel);
  XSetFillStyle(display, greyGC, FillSolid);

  lightGreyGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, lightGreyName, &availColor, &exactRGB);
  XSetForeground(display, lightGreyGC, availColor.pixel);
  XSetFillStyle(display, lightGreyGC, FillSolid);

  lightRedGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, lightRedName, &availColor, &exactRGB);
  XSetForeground(display, lightRedGC, availColor.pixel);
  XSetFillStyle(display, lightRedGC, FillSolid);

  dimmedRedGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, dimmedRedName, &availColor, &exactRGB);
  XSetForeground(display, dimmedRedGC, availColor.pixel);
  XSetFillStyle(display, dimmedRedGC, FillSolid);

  redGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, redName, &availColor, &exactRGB);
  XSetForeground(display, redGC, availColor.pixel);
  XSetFillStyle(display, redGC, FillSolid);

  lightGreenGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, lightGreenName, &availColor, &exactRGB);
  XSetForeground(display, lightGreenGC, availColor.pixel);
  XSetFillStyle(display, lightGreenGC, FillSolid);

  dimmedGreenGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, dimmedGreenName, &availColor, &exactRGB);
  XSetForeground(display, dimmedGreenGC, availColor.pixel);
  XSetFillStyle(display, dimmedGreenGC, FillSolid);

  greenGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, greenName, &availColor, &exactRGB);
  XSetForeground(display, greenGC, availColor.pixel);
  XSetFillStyle(display, greenGC, FillSolid);

  blackGC = XCreateGC(display, win, valuemask, &values);
  XAllocNamedColor(display, colormap, blackName, &availColor, &exactRGB);
  XSetForeground(display, blackGC, availColor.pixel);
  XSetFillStyle(display, blackGC, FillSolid);

  XMapWindow(display, win);

  // This call to XFlush must be here!  Without it, the first call
  // to an X Window drawing routine will fail (for some strange reason).
  XFlush(display);
}

// a polling call! not good; but ccj's main task is to keep track of the
// two player processes; not an interactive program really
void HandleXEvent()
{
  XEvent report;
  unsigned long eventMask = KeyPressMask | ButtonPressMask
    | ExposureMask | StructureNotifyMask;
  while (XCheckMaskEvent(display, eventMask, &report)) {
    switch (report.type) {
    case Expose:
      if (report.xexpose.count != 0)
	break;
      XClearWindow(display, win);
      DrawBoard();
      break;
    case ConfigureNotify:
      width = report.xconfigure.width;
      height = report.xconfigure.height;
      break;
    case ButtonPress:
      break;
    case KeyPress:
      XFreeGC(display, darkGreyGC);
      XFreeGC(display, greyGC);
      XFreeGC(display, lightGreyGC);
      XFreeGC(display, redGC);
      XFreeGC(display, greenGC);
      XFreeGC(display, blackGC);
      XCloseDisplay(display);
      KillPlayers();
      exit(0);
    default:
      break;
    }
  }
}










